@@ -1,28 +1,4 @@
-(******************************************************************************
- *                             Sexplib                                        *
- *                                                                            *
- * Copyright (C) 2005- Jane Street Holding, LLC                               *
- *    Contact: opensource@janestreet.com                                      *
- *    WWW: http://www.janestreet.com/ocaml                                    *
- *    Author: Markus Mottl                                                    *
- *                                                                            *
- * This library is free software; you can redistribute it and/or              *
- * modify it under the terms of the GNU Lesser General Public                 *
- * License as published by the Free Software Foundation; either               *
- * version 2 of the License, or (at your option) any later version.           *
- *                                                                            *
- * This library is distributed in the hope that it will be useful,            *
- * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
- * Lesser General Public License for more details.                            *
- *                                                                            *
- * You should have received a copy of the GNU Lesser General Public           *
- * License along with this library; if not, write to the Free Software        *
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
- *                                                                            *
- ******************************************************************************)
-
-(** Conv: Utility Module for S-expression Conversions *)
+(** Utility Module for S-expression Conversions *)
 
 open Bigarray
 
@@ -130,13 +106,13 @@ val sexp_of_ratio : Ratio.ratio -> Sexp.t
 (** [sexp_of_ratio n] converts the value [n] of type [Ratio.ratio] to an
     S-expression. *)
 
-val sexp_of_ref : ('a -> 'b) -> 'a ref -> 'b
+val sexp_of_ref : ('a -> Sexp.t) -> 'a ref -> Sexp.t
 (** [sexp_of_ref conv r] converts the value [r] of type ['a ref] to
     an S-expression.  Uses [conv] to convert values of type ['a] to an
     S-expression. *)
 
-val sexp_of_lazy : ('a -> 'b) -> 'a lazy_t -> 'b
-(** [sexp_of_ref conv l] converts the value [l] of type ['a lazy_t] to
+val sexp_of_lazy_t : ('a -> Sexp.t) -> 'a lazy_t -> Sexp.t
+(** [sexp_of_lazy_t conv l] converts the value [l] of type ['a lazy_t] to
     an S-expression.  Uses [conv] to convert values of type ['a] to an
     S-expression. *)
 
@@ -204,7 +180,7 @@ val sexp_of_opaque : 'a -> Sexp.t
     S-expression.  This means the user need not provide converters,
     but the result cannot be interpreted. *)
 
-val sexp_of_fun : _ -> Sexp.t
+val sexp_of_fun : ('a -> 'b) -> Sexp.t
 (** [sexp_of_fun f] converts the value [f] of function type to a
     dummy S-expression.  Functions cannot be serialized as S-expressions,
     but at least a placeholder can be generated for pretty-printing. *)
@@ -288,8 +264,8 @@ val ref_of_sexp : (Sexp.t -> 'a) -> Sexp.t -> 'a ref
     of type ['a ref] using conversion function [conv], which converts
     an S-expression to a value of type ['a]. *)
 
-val lazy_of_sexp : (Sexp.t -> 'a) -> Sexp.t -> 'a lazy_t
-(** [lazy_of_sexp conv sexp] converts S-expression [sexp] to a value
+val lazy_t_of_sexp : (Sexp.t -> 'a) -> Sexp.t -> 'a lazy_t
+(** [lazy_t_of_sexp conv sexp] converts S-expression [sexp] to a value
     of type ['a lazy_t] using conversion function [conv], which converts
     an S-expression to a value of type ['a]. *)
 
@@ -329,6 +305,10 @@ val hashtbl_of_sexp :
     key of type ['a], and function [conv_value], which converts an
     S-expression to hashtable value of type ['b]. *)
 
+val bigstring_of_sexp : Sexp.t -> bigstring
+(** [bigstring_of_sexp sexp] converts S-expression [sexp] to a
+    bigstring (character bigarray in C-layout). *)
+
 val float32_vec_of_sexp : Sexp.t -> float32_vec
 (** [float32_vec_of_sexp sexp] converts S-expression [sexp] to a
     one-dimensional bigarray of 32-bit floats in Fortran-layout. *)
@@ -355,7 +335,7 @@ val opaque_of_sexp : Sexp.t -> 'a
 (** [opaque_of_sexp sexp] @raise Of_sexp_error when attempting to
     convert an S-expression to an opaque value. *)
 
-val fun_of_sexp : Sexp.t -> ('a -> 'b)
+val fun_of_sexp : Sexp.t -> 'a
 (** [fun_of_sexp sexp] @raise Of_sexp_error when attempting to
     convert an S-expression to a function. *)
 
