@@ -1,27 +1,3 @@
-(******************************************************************************
- *                             Sexplib                                        *
- *                                                                            *
- * Copyright (C) 2005- Jane Street Holding, LLC                               *
- *    Contact: opensource@janestreet.com                                      *
- *    WWW: http://www.janestreet.com/ocaml                                    *
- *    Author: Markus Mottl                                                    *
- *                                                                            *
- * This library is free software; you can redistribute it and/or              *
- * modify it under the terms of the GNU Lesser General Public                 *
- * License as published by the Free Software Foundation; either               *
- * version 2 of the License, or (at your option) any later version.           *
- *                                                                            *
- * This library is distributed in the hope that it will be useful,            *
- * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
- * Lesser General Public License for more details.                            *
- *                                                                            *
- * You should have received a copy of the GNU Lesser General Public           *
- * License along with this library; if not, write to the Free Software        *
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
- *                                                                            *
- ******************************************************************************)
-
 (** Sexp_intf: interface specification for handling S-expressions (I/O, etc.) *)
 
 open Format
@@ -63,6 +39,14 @@ module type S = sig
       S-expressions from lex buffer [lexbuf] using the optional string
       buffer [buf] for storing intermediate strings. *)
 
+  val scan_rev_sexps : ?buf : Buffer.t -> Lexing.lexbuf -> t list
+  (** [scan_rev_sexps ?buf lexbuf] same as {!scan_sexps}, but returns the
+      reversed list and is slightly more efficient. *)
+
+  val scan_sexp_opt : ?buf : Buffer.t -> Lexing.lexbuf -> t option
+  (** [scan_sexp_opt ?buf lexbuf] is equivalent to [scan_sexp ?buf lexbuf]
+      except that it returns [None] when the eof is reached. *)
+
   val scan_iter_sexps :
     ?buf : Buffer.t -> f : (t -> unit) -> Lexing.lexbuf -> unit
   (** [scan_iter_sexps ?buf ~f lexbuf] iterates over all whitespace
@@ -115,16 +99,30 @@ module type S = sig
         [buf_pos] is set to [pos]. *)
   end
 
+  module Cont_state : sig
+    (** State of parser continuations *)
+    type t = Pre_sexp.Cont_state.t =
+      | Parsing_whitespace
+      | Parsing_atom
+      | Parsing_list
+      | Parsing_sexp_comment
+      | Parsing_block_comment
+
+    val to_string : t -> string
+    (** [to_string cont_state] converts state of parser continuation
+        [cont_state] to a string. *)
+  end
+
   (** Type of result from calling {!Sexp.parse}. *)
   type ('a, 't) parse_result = ('a, 't) Pre_sexp.parse_result =
     | Done of 't * Parse_pos.t  (** [Done (t, parse_pos)] finished parsing
                                     an S-expression.  Current parse position
                                     is [parse_pos]. *)
-    | Cont of bool * ('a, 't) parse_fun
-      (** [Cont (ws_only, parse_fun)] met the end of input before completely
+    | Cont of Cont_state.t * ('a, 't) parse_fun
+      (** [Cont (cont_state, parse_fun)] met the end of input before completely
           parsing an S-expression.  The user has to call [parse_fun] to
-          continue parsing the S-expression in another buffer.  If [ws_only]
-          is true, only whitespace has been parsed so far (or comments!).
+          continue parsing the S-expression in another buffer.  [cont_state]
+          is the current parsing state of the continuation.
           NOTE: the continuation may only be called once and will raise
           [Failure] otherwise! *)
 
@@ -279,12 +277,16 @@ module type S = sig
 
   val parse :
     ?parse_pos : Parse_pos.t -> ?len : int -> string -> (string, t) parse_result
-  (** [parse ?parse_pos ?len str] (partially) parses an S-expression in
-      string buffer [str] starting out with position information provided in
-      [parse_pos] and reading at most [len] characters.  To parse a single
-      atom that is not delimited by whitespace it is necessary to call this
-      function a second time with the returned continuation, and a dummy
-      buffer that contains whitespace.
+  (** [parse ?parse_pos ?len str] (partially) parses an S-expression in string buffer
+      [str] starting out with position information provided in [parse_pos] and reading at
+      most [len] characters.  To parse a single atom that is not delimited by whitespace
+      it is necessary to call this function a second time with the returned continuation,
+      and a dummy buffer that contains whitespace.
+
+      [parse] starts parsing [str] at position [parse_pos.buf_pos].  Each subsequent
+      [parse_fun] from a [Cont] uses the [buf] and [pos] that is supplied to it.  The
+      final [parse_fun] that returns [Done] mutates the [buf_pos] in the originally
+      supplied [parse_pos], and then returns it.
 
       @param parse_pos default = [Parse_pos.create ()]
       @param len default = [String.length str - parse_pos.Parse_pos.buf_pos]
@@ -554,6 +556,16 @@ module type S = sig
   val to_buffer : buf : Buffer.t -> t -> unit
   (** [to_buffer ~buf sexp] same as {!to_buffer_mach}. *)
 
+  val to_buffer_gen :
+       buf : 'buffer
+    -> add_char : ('buffer -> char -> unit)
+    -> add_string : ('buffer -> string -> unit)
+    -> t
+    -> unit
+  (** [to_buffer_gen ~buf ~add_char ~add_string sexp] outputs the S-expression
+      [sexp] converted to a string to buffer [buf] using the output functions
+      [add_char] and [add_string]. *)
+
 
   (** {6 Utilities for automated type conversions} *)
 
@@ -591,4 +603,63 @@ module type S = sig
   (** [subst_found sexp ~subst found] @return the S-expression that
       results from substituting [subst] within S-expression [sexp]
       at the location described by [found]. *)
+
+  (** S-expressions annotated with relative source positions and comments *)
+  module With_layout : sig
+
+    (* relative source positions *)
+    type pos = Src_pos.Relative.t = { row : int; col : int }
+    val sexp_of_pos : pos -> Type.t
+
+    (** S-expressions annotated with relative source positions and comments *)
+    type t =
+      | Atom of pos * string * string option (* second is quoted representation *)
+      | List of pos * t_or_comment list * pos (* positions of left and right parens *)
+    and t_or_comment =
+      | Sexp of t
+      | Comment of comment
+    and comment =
+      | Plain_comment of pos * string (* line or block comment *)
+      | Sexp_comment of pos * comment list * t (* position of #! *)
+
+    val sexp_of_t            : t -> Type.t
+    val sexp_of_comment      : comment -> Type.t
+    val sexp_of_t_or_comment : t_or_comment -> Type.t
+
+    module Forget : sig
+      val t             : t -> Type.t
+      val t_or_comment  : t_or_comment -> Type.t option
+      val t_or_comments : t_or_comment list -> Type.t list
+    end
+
+
+    module Render : sig
+      type asexp
+      type 'a t (* monad for position-respecting asexp rendering *)
+      val return : 'a -> 'a t
+      val bind : 'a t -> ('a -> 'b t) -> 'b t
+      val sexp : asexp -> unit t (* assumes that positions in [asexp] are relative *)
+      val run : (char -> unit) -> unit t -> unit
+    end
+      with type asexp := t_or_comment
+
+    module Parser : sig
+      type token
+      val sexp      : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> t_or_comment
+      val sexp_opt  : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> t_or_comment option
+      val sexps     : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> t_or_comment list
+      val rev_sexps : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> t_or_comment list
+      (* for debugging only, cannot be used otherwise anyway *)
+      val sexps_abs :
+        (Lexing.lexbuf -> token)
+        -> Lexing.lexbuf
+        -> Type_with_layout.Parsed.t_or_comment list
+    end
+
+    module Lexer : sig
+      val main : ?buf:Buffer.t -> Lexing.lexbuf -> Parser.token
+    end
+
+  end
+
 end
