@@ -1,35 +1,3 @@
-(******************************************************************************
- *                             Sexplib                                        *
- *                                                                            *
- * Copyright (C) 2005- Jane Street Holding, LLC                               *
- *    Contact: opensource@janestreet.com                                      *
- *    WWW: http://www.janestreet.com/ocaml                                    *
- *    Author: Markus Mottl                                                    *
- *                                                                            *
- * This file is derived from file "pa_tywith.ml" of version 0.45 of the       *
- * library "Tywith".                                                          *
- *                                                                            *
- * Tywith is Copyright (C) 2004, 2005 by                                      *
- *                                                                            *
- *    Martin Sandin  <msandin@hotmail.com>                                    *
- *                                                                            *
- *                                                                            *
- * This library is free software; you can redistribute it and/or              *
- * modify it under the terms of the GNU Lesser General Public                 *
- * License as published by the Free Software Foundation; either               *
- * version 2 of the License, or (at your option) any later version.           *
- *                                                                            *
- * This library is distributed in the hope that it will be useful,            *
- * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
- * Lesser General Public License for more details.                            *
- *                                                                            *
- * You should have received a copy of the GNU Lesser General Public           *
- * License along with this library; if not, write to the Free Software        *
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
- *                                                                            *
- ******************************************************************************)
-
 (* Pa_sexp_conv: Preprocessing Module for Automated S-expression Conversions *)
 
 open StdLabels
@@ -40,6 +8,8 @@ open Printf
 open Camlp4
 open PreCast
 
+open Syntax
+
 module Gen = Pa_type_conv.Gen
 
 (* Utility functions *)
@@ -63,6 +33,17 @@ let mk_rev_bindings loc fps =
   in
   bindings, patts, expr
 
+let mk_full_type loc type_name tps =
+  let coll_args tp _param = <:ctyp@loc< $tp$ _ >> in
+  List.fold_left ~f:coll_args ~init:<:ctyp@loc< $lid:type_name$ >> tps
+;;
+
+let sexp_type_is_recursive type_name tp =
+  Gen.type_is_recursive type_name tp
+    ~short_circuit:(function
+      | <:ctyp< sexp_opaque $_$ >> -> Some false
+      | _ -> None)
+
 let mk_bindings loc fps = mk_rev_bindings loc (List.rev fps)
 
 let unroll_cnv_fp loc var = function
@@ -83,12 +64,32 @@ let rec sig_of_tds cnv = function
       <:sig_item@loc< $sig_of_tds cnv tp1$; $sig_of_tds cnv tp2$ >>
   | _ -> assert false  (* impossible *)
 
+let type_app base_type types =
+  List.fold_left types ~init:base_type ~f:(fun acc typ ->
+    let loc = Ast.loc_of_ctyp typ in
+    <:ctyp@loc< $acc$ $typ$ >>)
+;;
+
+(* Generates the quantified type [ ! 'a .. 'z . (make_mono_type t ('a .. 'z)) ], or [None]
+   if there are no type parameters.
+
+   A quantified type annotation is required for of_sexp and to_sexp functions for
+   non-regular types, such as [ type t 'a = [ A of 'a | B of t 'a 'a ] ]. *)
+let mk_poly_type make_mono_type loc type_name type_params =
+  let type_params = List.map type_params ~f:Gen.drop_variance_annotations in
+  match type_params with
+  | [] -> None
+  | first :: rest ->
+    let unquantified_type = make_mono_type <:ctyp@loc< $lid:type_name$ >> type_params in
+    let loc = Ast.loc_of_ctyp unquantified_type in
+    (* It's confusing that you need an application between the '!' and the '.'. I would
+       have expected a list. *)
+    Some <:ctyp@loc< ! $type_app first rest$ . $unquantified_type$ >>
 
 (* Generators for S-expressions *)
 
 (* Generates the signature for type conversion to S-expressions *)
 module Sig_generate_sexp_of = struct
-
   let rec sig_of_td__loop acc = function
     | [] ->
         let loc = Ast.loc_of_ctyp acc in
@@ -103,16 +104,16 @@ module Sig_generate_sexp_of = struct
     let sexp_of = sig_of_td__loop <:ctyp@loc< $lid:type_name$ >> tps in
     <:sig_item@loc< value $lid: "sexp_of_" ^ type_name$ : $sexp_of$ >>
 
-  let mk_sig tds = <:sig_item< $sig_of_tds sig_of_td tds$ >>
+  let mk_sig _rec tds = <:sig_item< $sig_of_tds sig_of_td tds$ >>
 
-  let () = Pa_type_conv.add_sig_generator "sexp_of" mk_sig
+  let () = Pa_type_conv.add_sig_generator ~delayed:true "sexp_of" mk_sig
 
-  let mk_sig_exn = function
+  let mk_sig_exn _rec = function
     | <:ctyp@loc< $uid:_$ >> | <:ctyp@loc< $uid:_$ of $_$ >> ->
         <:sig_item@loc< >>
     | tp -> Gen.error tp ~fn:"mk_sig_exn" ~msg:"unknown type"
 
-  let () = Pa_type_conv.add_sig_generator ~is_exn:true "sexp" mk_sig_exn
+  let () = Pa_type_conv.add_sig_generator ~delayed:true ~is_exn:true "sexp" mk_sig_exn
 end
 
 
@@ -120,6 +121,7 @@ end
 module Sig_generate_of_sexp = struct
 
   let rec is_polymorphic_variant = function
+    | <:ctyp< ($ty$ as $_$) >> -> is_polymorphic_variant ty
     | <:ctyp< private $tp$ >> -> is_polymorphic_variant tp
     | <:ctyp< ( $tup:_$ ) >>
     | <:ctyp< $_$ -> $_$ >>
@@ -158,42 +160,72 @@ module Sig_generate_of_sexp = struct
           ~msg:"sexp_poly annotation \
             but type is surely not a polymorphic variant"
     | false, (`Surely_not | `Maybe) -> of_sexp_item
-    | (true | false), `Definitely | true, `Maybe ->
+    | (true | false), `Definitely
+    | true, `Maybe ->
         <:sig_item@loc<
           $of_sexp_item$;
-          value $lid: type_name ^ "_of_sexp__"$ : $of_sexp$;
+          value $lid: "__" ^ type_name ^ "_of_sexp__"$ : $of_sexp$;
         >>
 
-  let mk_sig with_poly tds =
+  let mk_sig with_poly _rec tds =
     <:sig_item< $sig_of_tds (sig_of_td with_poly) tds$ >>
 
-  let () = Pa_type_conv.add_sig_generator "of_sexp" (mk_sig false)
-  let () = Pa_type_conv.add_sig_generator "of_sexp_poly" (mk_sig true)
+  let () = Pa_type_conv.add_sig_generator ~delayed:true "of_sexp" (mk_sig false)
+  let () = Pa_type_conv.add_sig_generator ~delayed:true "of_sexp_poly" (mk_sig true)
 end
 
 
 (* Generates the signature for type conversion to S-expressions *)
 module Sig_generate = struct
   let () =
-    Pa_type_conv.add_sig_generator "sexp" (fun tds ->
-      let loc = Ast.loc_of_ctyp tds in
-      <:sig_item@loc<
-        $Sig_generate_sexp_of.mk_sig tds$;
-        $Sig_generate_of_sexp.mk_sig false tds$
-      >>)
+    Pa_type_conv.add_sig_set
+      "sexp"
+      ~set:["sexp_of"; "of_sexp"]
 
   let () =
-    Pa_type_conv.add_sig_generator "sexp_poly" (fun tds ->
-      let loc = Ast.loc_of_ctyp tds in
-      <:sig_item@loc<
-        $Sig_generate_sexp_of.mk_sig tds$;
-        $Sig_generate_of_sexp.mk_sig true tds$
-      >>)
+    Pa_type_conv.add_sig_set
+      "sexp_poly"
+      ~set:["sexp_of"; "of_sexp_poly"]
 end
 
 
 (* Generator for converters of OCaml-values to S-expressions *)
 module Generate_sexp_of = struct
+  (* Handling of record defaults *)
+
+  type record_field_handler = [ `keep | `drop_default | `drop_if of Ast.expr ]
+
+  let record_field_handlers
+      : (Loc.t, record_field_handler) Hashtbl.t
+      = Hashtbl.create 0
+
+  let get_record_field_handler loc =
+    try Hashtbl.find record_field_handlers loc
+    with Not_found -> `keep
+
+  let check_record_field_handler loc =
+    if Hashtbl.mem record_field_handlers loc then
+      Loc.raise loc (Failure "sexp record field handler defined twice")
+
+  let () =
+    Pa_type_conv.add_record_field_generator "sexp_drop_default" (fun tp ->
+      let loc = Ast.loc_of_ctyp tp in
+      check_record_field_handler loc;
+      Hashtbl.replace record_field_handlers ~key:loc ~data:`drop_default)
+
+  let () =
+    Pa_type_conv.add_record_field_generator_with_arg "sexp_drop_if"
+      Syntax.expr (fun expr_opt tp ->
+        let loc = Ast.loc_of_ctyp tp in
+        check_record_field_handler loc;
+        let test =
+          match expr_opt with
+          | Some expr -> expr
+          | None -> Loc.raise loc (Failure "could not parse expression")
+        in
+        Hashtbl.replace record_field_handlers ~key:loc ~data:(`drop_if test))
+
+  (* Make abstract calls *)
   let mk_abst_call loc tn rev_path =
     <:expr@loc<
       $id:Gen.ident_of_rev_path loc (("sexp_of_" ^ tn) :: rev_path)$
@@ -207,14 +239,16 @@ module Generate_sexp_of = struct
 
   (* Conversion of types *)
   let rec sexp_of_type = function
+    | <:ctyp@loc< _ >> ->
+        `Fun <:expr@loc< fun _ -> Sexp.Atom "_" >>
     | <:ctyp@loc< sexp_opaque $_$ >> ->
         `Fun <:expr@loc< Sexplib.Conv.sexp_of_opaque >>
     | <:ctyp@loc< $tp1$ $tp2$ >> -> `Fun (sexp_of_appl_fun loc tp1 tp2)
     | <:ctyp< ( $tup:tp$ ) >> -> sexp_of_tuple tp
     | <:ctyp@loc< '$parm$ >> -> `Fun <:expr@loc< $lid:"_of_" ^ parm$ >>
     | <:ctyp@loc< $id:id$ >> -> `Fun (sexp_of_path_fun loc id)
-    | <:ctyp@loc< $_$ -> $_$ >> as arrow ->
-        `Fun <:expr@loc< fun (_f : $arrow$) ->
+    | <:ctyp@loc< $_$ -> $_$ >> ->
+        `Fun <:expr@loc< fun _f ->
           Sexplib.Conv.sexp_of_fun Pervasives.ignore >>
     | <:ctyp< [< $row_fields$ ] >> | <:ctyp< [> $row_fields$ ] >>
     | <:ctyp< [= $row_fields$ ] >> -> sexp_of_variant row_fields
@@ -352,6 +386,8 @@ module Generate_sexp_of = struct
             let $bindings$ in
             Sexplib.Sexp.List $Gen.mk_expr_lst loc (cnstr_expr :: vars)$
         >>
+    | <:ctyp< $_$ : $_$ >> as tp -> Gen.error tp ~fn:"branch_sum"
+      ~msg:"GADTs are not supported by sexplib"
     | tp -> Gen.unknown_type tp "branch_sum"
 
   let sexp_of_sum alts = `Match (branch_sum alts)
@@ -363,7 +399,7 @@ module Generate_sexp_of = struct
     let p = <:patt@loc< $lid:name$ = $lid:"v_" ^ name$ >> in
     <:patt@loc< $patt$; $p$ >>
 
-  let sexp_of_default_field patt expr name tp sexp_of empty =
+  let sexp_of_record_field patt expr name tp ?sexp_of is_empty_expr =
     let loc = Ast.loc_of_ctyp tp in
     let patt = mk_rec_patt loc patt name in
     let cnv_expr =
@@ -372,84 +408,120 @@ module Generate_sexp_of = struct
       | `Match matchings ->
           <:expr@loc< fun el -> match el with [ $matchings$ ] >>
     in
+    let cnv_expr =
+      match sexp_of with
+      | None -> cnv_expr
+      | Some sexp_of -> <:expr@loc< $sexp_of$ $cnv_expr$ >>
+    in
     let expr =
       let v_name = <:expr@loc< $lid: "v_" ^ name$ >> in
       <:expr@loc<
         let bnds =
-          if $v_name$ = $empty$ then bnds
+          if $is_empty_expr loc v_name$ then bnds
           else
-            let arg = $sexp_of$ $cnv_expr$ $v_name$ in
+            let arg = $cnv_expr$ $v_name$ in
             let bnd =
               Sexplib.Sexp.List [Sexplib.Sexp.Atom $str:name$; arg]
             in
             [ bnd :: bnds ]
         in
         $expr$
-      >>
+        >>
     in
     patt, expr
 
+  let sexp_of_default_field patt expr name tp ?sexp_of default =
+    sexp_of_record_field patt expr name tp ?sexp_of
+      (fun loc expr -> <:expr@loc< Pervasives.(=) $default$ $expr$ >>)
+
   let sexp_of_record flds_ctyp =
     let flds = Ast.list_of_ctyp flds_ctyp [] in
-    let rec coll (patt, expr) = function
+    let list_empty_expr loc lst = <:expr@loc<
+      match $lst$ with
+        [ [] -> True
+        | _ -> False
+        ]
+        >>
+    in
+    let array_empty_expr loc arr = <:expr@loc<
+      match $arr$ with
+        [ [||] -> True
+        | _ -> False
+        ]
+        >>
+    in
+    let coll (patt, expr) = function
       | <:ctyp@loc< $lid:name$ : mutable sexp_option $tp$ >>
       | <:ctyp@loc< $lid:name$ : sexp_option $tp$ >> ->
-          let patt = mk_rec_patt loc patt name in
-          let vname = <:expr@loc< v >> in
-          let cnv_expr = unroll_cnv_fp loc vname (sexp_of_type tp) in
-          let expr =
-            <:expr@loc<
-              let bnds =
-                match $lid:"v_" ^ name$ with
+        let patt = mk_rec_patt loc patt name in
+        let vname = <:expr@loc< v >> in
+        let cnv_expr = unroll_cnv_fp loc vname (sexp_of_type tp) in
+        let expr =
+          <:expr@loc<
+            let bnds =
+              match $lid:"v_" ^ name$ with
                 [ None -> bnds
                 | Some v ->
-                    let arg = $cnv_expr$ in
-                    let bnd =
-                      Sexplib.Sexp.List [Sexplib.Sexp.Atom $str:name$; arg]
-                    in
-                    [ bnd :: bnds ] ]
-              in
-              $expr$
+                  let arg = $cnv_expr$ in
+                  let bnd =
+                    Sexplib.Sexp.List [Sexplib.Sexp.Atom $str:name$; arg]
+                  in
+                  [ bnd :: bnds ] ]
+            in
+            $expr$
             >>
-          in
-          patt, expr
+        in
+        patt, expr
       | <:ctyp@loc< $lid:name$ : mutable sexp_bool >>
       | <:ctyp@loc< $lid:name$ : sexp_bool >> ->
-          let patt = mk_rec_patt loc patt name in
-          let expr =
-            <:expr@loc<
-              let bnds =
-                if $lid:"v_" ^ name$ then
-                  let bnd = Sexplib.Sexp.List [Sexplib.Sexp.Atom $str:name$] in
-                  [ bnd :: bnds ]
-                else bnds
-              in
-              $expr$
+        let patt = mk_rec_patt loc patt name in
+        let expr =
+          <:expr@loc<
+            let bnds =
+              if $lid:"v_" ^ name$ then
+                let bnd = Sexplib.Sexp.List [Sexplib.Sexp.Atom $str:name$] in
+                [ bnd :: bnds ]
+              else bnds
+            in
+            $expr$
             >>
-          in
-          patt, expr
+        in
+        patt, expr
       | <:ctyp@loc< $lid:name$ : mutable sexp_list $tp$ >>
       | <:ctyp@loc< $lid:name$ : sexp_list $tp$ >> ->
-          sexp_of_default_field
-            patt expr name tp <:expr@loc< sexp_of_list >> <:expr@loc< [] >>
+        sexp_of_record_field patt expr name tp
+          ~sexp_of:<:expr@loc< sexp_of_list >> list_empty_expr
       | <:ctyp@loc< $lid:name$ : mutable sexp_array $tp$ >>
       | <:ctyp@loc< $lid:name$ : sexp_array $tp$ >> ->
-          sexp_of_default_field
-            patt expr name tp <:expr@loc< sexp_of_array >> <:expr@loc< [||] >>
+        sexp_of_record_field patt expr name tp
+          ~sexp_of:<:expr@loc< sexp_of_array >> array_empty_expr
       | <:ctyp@loc< $lid:name$ : mutable $tp$ >>
       | <:ctyp@loc< $lid:name$ : $tp$ >> ->
-          let patt = mk_rec_patt loc patt name in
-          let vname = <:expr@loc< $lid:"v_" ^ name$ >> in
-          let cnv_expr = unroll_cnv_fp loc vname  (sexp_of_type tp) in
-          let expr =
-            <:expr@loc<
-              let arg = $cnv_expr$ in
-              let bnd = Sexplib.Sexp.List [Sexplib.Sexp.Atom $str:name$; arg] in
-              let bnds = [ bnd :: bnds ] in
-              $expr$
-            >>
-          in
-          patt, expr
+        let opt_default = Pa_type_conv.Gen.find_record_default loc in
+        let field_handler = get_record_field_handler loc in
+        begin match opt_default, field_handler with
+        | None, `drop_default -> Loc.raise loc (Failure "no default to drop")
+        | _, `drop_if test ->
+          sexp_of_record_field patt expr name tp
+              (fun loc expr -> <:expr@loc< $test$ $expr$>>)
+        | Some default, `drop_default ->
+          sexp_of_default_field patt expr name tp default
+        | _, `keep ->
+            let patt = mk_rec_patt loc patt name in
+            let vname = <:expr@loc< $lid:"v_" ^ name$ >> in
+            let cnv_expr = unroll_cnv_fp loc vname (sexp_of_type tp) in
+            let expr =
+              <:expr@loc<
+                let arg = $cnv_expr$ in
+                let bnd =
+                  Sexplib.Sexp.List [Sexplib.Sexp.Atom $str:name$; arg]
+                in
+                let bnds = [ bnd :: bnds ] in
+                $expr$
+              >>
+            in
+            patt, expr
+        end
       | _ -> assert false  (* impossible *)
     in
     let loc = Ast.loc_of_ctyp flds_ctyp in
@@ -487,15 +559,20 @@ module Generate_sexp_of = struct
           (* Prevent violation of value restriction and problems with
              recursive types by eta-expanding function definitions *)
           <:expr@loc< fun [ v -> $fun_expr$ v ] >>
-      | `Match matchings -> <:expr@loc< fun [ $matchings$ ] >>
+      | `Match matchings ->
+          <:expr@loc< fun [ $matchings$ ] >>
     in
-    let mk_pat id = <:patt@loc< $lid:id$ >> in
     let patts =
       List.map tps
         ~f:(fun ty -> <:patt@loc< $lid:"_of_" ^ Gen.get_tparam_id ty$>>)
     in
-    let bnd = mk_pat ("sexp_of_" ^ type_name) in
-    <:binding@loc< $bnd$ = $Gen.abstract loc patts body$ >>
+    let body = Gen.abstract loc patts body in
+    let body =
+      match mk_poly_type Sig_generate_sexp_of.sig_of_td__loop loc type_name tps with
+      | None -> body
+      | Some typ -> <:expr@loc< ( $body$ : $typ$ ) >>
+    in
+    <:binding@loc< $lid:"sexp_of_" ^ type_name$ = $body$ >>
 
   let rec sexp_of_tds = function
     | Ast.TyDcl (loc, type_name, tps, rhs, _cl) ->
@@ -504,13 +581,13 @@ module Generate_sexp_of = struct
         <:binding@loc< $sexp_of_tds tp1$ and $sexp_of_tds tp2$ >>
     | _ -> assert false  (* impossible *)
 
-  let sexp_of tds =
+  let sexp_of rec_ tds =
     let binding, recursive, loc =
       match tds with
       | Ast.TyDcl (loc, type_name, tps, rhs, _cl) ->
           sexp_of_td loc type_name tps rhs,
-          Gen.type_is_recursive type_name rhs, loc
-      | Ast.TyAnd (loc, _, _) as tds -> sexp_of_tds tds, true, loc
+          rec_ && sexp_type_is_recursive type_name rhs, loc
+      | Ast.TyAnd (loc, _, _) as tds -> sexp_of_tds tds, rec_, loc
       | _ -> assert false  (* impossible *)
     in
     if recursive then <:str_item@loc< value rec $binding$ >>
@@ -519,11 +596,7 @@ module Generate_sexp_of = struct
   (* Add code generator to the set of known generators *)
   let () = Pa_type_conv.add_generator "sexp_of" sexp_of
 
-  let string_of_ident id =
-    let str_lst = Gen.get_rev_id_path id [] in
-    String.concat ~sep:"." str_lst
-
-  let sexp_of_exn tp =
+  let sexp_of_exn _rec tp =
     let get_full_cnstr cnstr = Pa_type_conv.get_conv_path () ^ "." ^ cnstr in
     let expr =
       match tp with
@@ -572,7 +645,7 @@ end
 module Generate_of_sexp = struct
   let mk_abst_call loc tn ?(internal = false) rev_path =
     let tns = tn ^ "_of_sexp" in
-    let tns_suff = if internal then tns ^ "__" else tns in
+    let tns_suff = if internal then "__" ^ tns ^ "__" else tns in
     <:expr@loc< $id:Gen.ident_of_rev_path loc (tns_suff :: rev_path)$ >>
 
   (* Utility functions for polymorphic variants *)
@@ -665,9 +738,7 @@ module Generate_of_sexp = struct
     | <:ctyp< ( $tup:tp$ ) >> -> tuple_of_sexp tp
     | <:ctyp@loc< '$parm$ >> -> `Fun <:expr@loc< $lid:"_of_" ^ parm$ >>
     | <:ctyp@loc< $id:id$ >> -> `Fun (path_of_sexp_fun loc id)
-    | <:ctyp@loc< $_$ -> $_$ >> as arrow ->
-        `Fun <:expr@loc< fun sexp ->
-          (Sexplib.Conv.fun_of_sexp sexp : $arrow$) >>
+    | <:ctyp@loc< $_$ -> $_$ >> -> `Fun <:expr@loc< Sexplib.Conv.fun_of_sexp >>
     | <:ctyp< [< $row_fields$ ] >> | <:ctyp< [> $row_fields$ ] >>
     | <:ctyp< [= $row_fields$ ] >> ->
         variant_of_sexp ?full_type:None row_fields
@@ -730,7 +801,7 @@ module Generate_of_sexp = struct
           handle_variant_inh full_type match_last other_matches inh
     in
     let other_matches =
-      mk_variant_other_matches loc rev_structs "ptag_no_args"
+      mk_variant_other_matches loc rev_structs "ptag_takes_args"
     in
     let match_atoms_inhs, match_last =
       List.fold_left ~f:coll ~init:(other_matches, false) rev_atoms_inhs in
@@ -827,8 +898,15 @@ module Generate_of_sexp = struct
 
   (* Generate matching code for variants *)
   and variant_of_sexp ?full_type row_tp =
+    let rec replace_params_with_underscores = function
+      | <:ctyp@loc< $a$ $_$ >> -> <:ctyp@loc< $replace_params_with_underscores a$ _ >>
+      | x -> x
+    in
     let loc = Ast.loc_of_ctyp row_tp in
     let row_fields = Ast.list_of_ctyp row_tp [] in
+    let row_tp =
+      Ast.tyOr_of_list (List.map row_fields ~f:replace_params_with_underscores)
+    in
     let is_contained, full_type =
       match full_type with
       | None -> true, <:ctyp@loc< [= $row_tp$ ] >>
@@ -1019,24 +1097,28 @@ module Generate_of_sexp = struct
       let rec loop (res_tpls, bi_lst, good_patts as acc) = function
         | <:ctyp@loc< $lid:nm$ : $tp$ >> ->
             let fld = <:expr@loc< $lid:nm ^ "_field"$.val >> in
+            let mk_default loc =
+              bi_lst, <:patt@loc< $lid:nm ^ "_value"$ >> :: good_patts
+            in
             let new_bi_lst, new_good_patts =
               match tp with
-              | <:ctyp@loc< sexp_bool >> |  <:ctyp@loc< mutable sexp_bool >>
+              | <:ctyp@loc< sexp_bool >> | <:ctyp@loc< mutable sexp_bool >>
               | <:ctyp@loc< sexp_option $_$ >>
               | <:ctyp@loc< mutable sexp_option $_$ >>
               | <:ctyp@loc< sexp_list $_$ >>
               | <:ctyp@loc< mutable sexp_list $_$ >>
               | <:ctyp@loc< sexp_array $_$ >>
-              | <:ctyp@loc< mutable sexp_array $_$ >> ->
-                  bi_lst, <:patt@loc< $lid:nm ^ "_value"$ >> :: good_patts
-              | _ ->
-                  let loc = Ast.loc_of_ctyp tp in
-                  has_nonopt_fields := true;
-                  (
-                    <:expr@loc<
-                      (Pervasives.(=) $fld$ None, $str:nm$) >> :: bi_lst,
-                    <:patt@loc< Some $lid:nm ^ "_value"$ >> :: good_patts
-                  )
+              | <:ctyp@loc< mutable sexp_array $_$ >> -> mk_default loc
+              | <:ctyp@loc< $_$ >> ->
+                  match Pa_type_conv.Gen.find_record_default loc with
+                  | Some _ -> mk_default loc
+                  | None ->
+                      has_nonopt_fields := true;
+                      (
+                        <:expr@loc<
+                          (Pervasives.(=) $fld$ None, $str:nm$) >> :: bi_lst,
+                        <:patt@loc< Some $lid:nm ^ "_value"$ >> :: good_patts
+                      )
             in
             (
               <:expr@loc< $fld$ >> :: res_tpls,
@@ -1065,20 +1147,31 @@ module Generate_of_sexp = struct
         let rec loop = function
           | <:ctyp@loc< $tp1$; $tp2$ >> ->
               <:rec_binding@loc< $loop tp1$; $loop tp2$ >>
+          | <:ctyp@loc< $lid:nm$ : mutable sexp_list $_$ >>
           | <:ctyp@loc< $lid:nm$ : sexp_list $_$ >> ->
               <:rec_binding@loc<
                 $lid:nm$ =
                   match $lid:nm ^ "_value"$ with
                   [ None -> [] | Some v -> v ]
               >>
+          | <:ctyp@loc< $lid:nm$ : mutable sexp_array $_$ >>
           | <:ctyp@loc< $lid:nm$ : sexp_array $_$ >> ->
               <:rec_binding@loc<
                 $lid:nm$ =
                   match $lid:nm ^ "_value"$ with
                   [ None -> [||] | Some v -> v ]
               >>
+          | <:ctyp@loc< $lid:nm$ : mutable $_$ >>
           | <:ctyp@loc< $lid:nm$ : $_$ >> ->
-              <:rec_binding@loc< $lid:nm$ = $lid:nm ^ "_value"$ >>
+              begin match Pa_type_conv.Gen.find_record_default loc with
+              | None -> <:rec_binding@loc< $lid:nm$ = $lid:nm ^ "_value"$ >>
+              | Some default ->
+                  <:rec_binding@loc<
+                    $lid:nm$ =
+                      match $lid:nm ^ "_value"$ with
+                      [ None -> $default$ | Some v -> v ]
+                  >>
+              end
           | _ -> assert false  (* impossible *)
         in
         <:expr@loc< { $loop flds$ } >>
@@ -1134,7 +1227,7 @@ module Generate_of_sexp = struct
               match field_name with
               [ $mc_no_args_fields$ ];
               iter tail }
-        | [sexp :: _] ->
+        | [((Sexplib.Sexp.Atom _ | Sexplib.Sexp.List _) as sexp) :: _] ->
             Sexplib.Conv_error.record_only_pairs_expected _tp_loc sexp
         | [] -> () ]
       in
@@ -1205,23 +1298,16 @@ module Generate_of_sexp = struct
 
   (* Generate code from type definitions *)
 
-  let rec is_poly_call = function
-    | <:expr< $f$ $_$ >> -> is_poly_call f
-    | <:expr< $lid:name$ >> -> name.[0] = '_' && name.[1] = 'o'
-    | _ -> false
-
   let td_of_sexp loc type_name tps rhs =
-    let is_alias_ref = ref false in
+    let alias_ref = ref `Not_an_alias in
     let handle_alias tp =
-      is_alias_ref := true;
+      alias_ref :=
+        (match tp with
+        | <:ctyp< '$_$ >> -> `Alias `Type_var
+        | _ -> `Alias `Type_constructor);
       type_of_sexp tp
     in
-    let coll_args tp param =
-      <:ctyp@loc< $tp$ $Gen.drop_variance_annotations param$ >>
-    in
-    let full_type =
-      List.fold_left ~f:coll_args ~init:<:ctyp@loc< $lid:type_name$ >> tps
-    in
+    let full_type = mk_full_type loc type_name tps in
     let is_variant_ref = ref false in
     let handle_variant row_fields =
       is_variant_ref := true;
@@ -1244,7 +1330,8 @@ module Generate_of_sexp = struct
           <:expr@loc< fun [ t -> $fun_expr$ t ] >>
       | `Match matchings -> <:expr@loc< fun [ $matchings$ ] >>
     in
-    let internal_name = type_name ^ "_of_sexp__" in
+    let external_name = type_name ^ "_of_sexp" in
+    let internal_name = "__" ^ type_name ^ "_of_sexp__" in
     let arg_patts, arg_exprs =
       List.split (
         List.map ~f:(function tp ->
@@ -1253,12 +1340,25 @@ module Generate_of_sexp = struct
           )
           tps)
     in
-    let with_poly_call = !is_alias_ref && is_poly_call body in
+    let with_poly_call =
+      match !alias_ref with
+      | `Not_an_alias
+      | `Alias `Type_constructor -> false
+      | `Alias `Type_var -> true in
     let internal_fun_body =
       let full_type_name =
         sprintf "%s.%s" (Pa_type_conv.get_conv_path ()) type_name
       in
       if with_poly_call then
+        (* special case for type definitions whose bodies are type variables, like:
+             type ('a, 'b) t = 'a
+           because
+           - they can used in polymorphic variants: [ ([`A], int) t | `B ]
+           - the way sexplib works, it cannot handle backtracking in these cases,
+             (because we only receive as parameter sexp_of_'a but not
+             __sexp_of_'a__ presumably)
+             so it is better to emit an error rather than do something weird
+        *)
         Gen.abstract loc arg_patts
           <:expr@loc<
             fun sexp ->
@@ -1284,21 +1384,26 @@ module Generate_of_sexp = struct
       if with_poly_call then
         <:expr@loc< try $body$ sexp with [ $no_variant_match_mc$ ] >>
       (* Type alias may refer to variant, therefore same handling here! *)
-      else if !is_variant_ref || !is_alias_ref then
+      else if !is_variant_ref || !alias_ref = `Alias `Type_constructor then
         <:expr@loc< try $internal_call$ with [ $no_variant_match_mc$ ] >>
       else internal_call
     in
-    let internal_binding =
-      <:binding@loc< $lid:internal_name$ = $internal_fun_body$ >>
-    in
-    let external_fun_patt = <:patt@loc< $lid:type_name ^ "_of_sexp"$ >> in
     let external_fun_body =
       Gen.abstract loc arg_patts
-        <:expr@loc< fun sexp -> $pre_external_fun_body$ >>
+        <:expr@loc< fun sexp -> ($pre_external_fun_body$) >>
     in
-    let external_binding =
-      <:binding@loc< $external_fun_patt$ = $external_fun_body$ >>
+    let type_of_of_sexp =
+      mk_poly_type Sig_generate_of_sexp.sig_of_td__loop loc type_name tps
     in
+    let maybe_annotate expr =
+      match type_of_of_sexp with
+      | None -> expr
+      | Some typ -> <:expr@loc< ( $expr$ : $typ$ ) >>
+    in
+    let internal_fun_body = maybe_annotate internal_fun_body in
+    let external_fun_body = maybe_annotate external_fun_body in
+    let internal_binding = <:binding@loc< $lid:internal_name$ = $internal_fun_body$ >> in
+    let external_binding = <:binding@loc< $lid:external_name$ = $external_fun_body$ >> in
     internal_binding, external_binding
 
   let rec tds_of_sexp acc = function
@@ -1308,21 +1413,21 @@ module Generate_of_sexp = struct
     | _ -> assert false  (* impossible *)
 
   (* Generate code from type definitions *)
-  let of_sexp = function
+  let of_sexp rec_ = function
     | Ast.TyDcl (loc, type_name, tps, rhs, _cl) ->
         let internal_binding, external_binding =
           td_of_sexp loc type_name tps rhs
         in
-        let recursive = Gen.type_is_recursive type_name rhs in
-        if recursive then
+        let is_recursive = rec_ && sexp_type_is_recursive type_name rhs in
+        if is_recursive then
           <:str_item@loc<
             value rec $internal_binding$
-            and $external_binding$
+            and $external_binding$;
           >>
         else
           <:str_item@loc<
             value $internal_binding$;
-            value $external_binding$
+            value $external_binding$;
           >>
     | Ast.TyAnd (loc, _, _) as tds ->
         let two_bindings = tds_of_sexp [] tds in
@@ -1330,7 +1435,14 @@ module Generate_of_sexp = struct
           List.map ~f:(fun (b1, b2) -> <:binding@loc< $b1$ and $b2$ >>)
             two_bindings
         in
-        <:str_item@loc< value rec $list:bindings$ >>
+        if rec_ then
+          <:str_item@loc<
+            value rec $list:bindings$;
+          >>
+        else
+          <:str_item@loc<
+            value $list:bindings$;
+          >>
     | _ -> assert false  (* impossible *)
 
   (* Add code generator to the set of known generators *)
@@ -1340,12 +1452,12 @@ end
 module Quotations = struct
   let of_sexp_quote loc _loc_name_opt cnt_str =
     Pa_type_conv.set_conv_path_if_not_set loc;
-    let ctyp = Gram.parse_string Syntax.ctyp_quot loc cnt_str in
+    let ctyp = Gram.parse_string ctyp_quot loc cnt_str in
     let fp = Generate_of_sexp.type_of_sexp ctyp in
     let body =
       match fp with
       | `Fun fun_expr -> <:expr@loc< $fun_expr$ sexp >>
-      | `Match matchings -> <:expr@loc< match sexp with [$matchings$]  >>
+      | `Match matchings -> <:expr@loc< match sexp with [$matchings$] >>
     in
     let full_type_name =
       sprintf "%s line %i: %s"
@@ -1358,26 +1470,18 @@ module Quotations = struct
       >>
 
   let () =
-    Syntax.Quotation.add "of_sexp" Syntax.Quotation.DynAst.expr_tag
-      of_sexp_quote
+    Quotation.add "of_sexp" Quotation.DynAst.expr_tag of_sexp_quote
 
- let sexp_of_quote loc _loc_name_opt cnt_str =
-   Pa_type_conv.set_conv_path_if_not_set loc;
-   let ctyp = Gram.parse_string Syntax.ctyp_quot loc cnt_str in
-   Generate_sexp_of.mk_cnv_expr ctyp
+  let sexp_of_quote loc _loc_name_opt cnt_str =
+    Pa_type_conv.set_conv_path_if_not_set loc;
+    let ctyp = Gram.parse_string ctyp_quot loc cnt_str in
+    Generate_sexp_of.mk_cnv_expr ctyp
 
-  let () =
-    Syntax.Quotation.add "sexp_of" Syntax.Quotation.DynAst.expr_tag
-      sexp_of_quote
+  let () = Quotation.add "sexp_of" Quotation.DynAst.expr_tag sexp_of_quote
 end
 
 (* Add "of_sexp" and "sexp_of" as "sexp" to the set of generators *)
 let () =
-  Pa_type_conv.add_generator
+  Pa_type_conv.add_str_set
     "sexp"
-    (fun tds ->
-      let loc = Ast.loc_of_ctyp tds in
-      <:str_item@loc<
-        $Generate_of_sexp.of_sexp tds$; $Generate_sexp_of.sexp_of tds$
-      >>
-    )
+    ~set:["of_sexp"; "sexp_of"]
